module GA(
  GAData,
  dataRange,
  repLength,
  fitness,
  maxFitness,
  showSol,
  crossover,
  mutate,
  initialize
) where

import Data.List.Grouping(splitEvery)

class GAData a where
  dataRange :: a -> (Int, Int)
  repLength :: a -> Int
  fitness :: a -> [Int] -> Int
  maxFitness :: a -> Int
  showSol :: a -> [Int] -> [Char]


-- These functions will be common to all genetic instances
initialize :: GAData a => a -> [Int] -> Int -> ([[Int]], [Int])
initialize d l popSize = (splitEvery r vals, newVals)
  where
    r = repLength d
    (vals, newVals) = splitAt (r*popSize) l



{-
      The following will be implemented as Uniform Crossover:
      We could generate a random number to see which int is kept in the child.
      We will use splitting instead.
      References:
      How to generate different pcts
      https://mail.haskell.org/pipermail/beginners/2010-April/004058.html
-}
crossover :: GAData a => a -> ([Int], [Int]) -> [Int]
crossover d (a, b) = take half a ++ drop half b
    where (half, _) = quotRem (repLength d) 2


mutate :: GAData a => a -> Float -> [Float] -> [Int] -> [Int] -> [Int]
mutate _ mutPct pcts rvs sol = map keep $ zip3 pcts rvs sol
  where
    keep (r, x', x)
      | r > mutPct = x
      | otherwise = x'
