module TSP(
  TSP(TSP)
) where

import GA
import qualified Data.Map as M
import Data.List.Unique(allUnique)
import Data.Maybe(fromMaybe)

data TSP = TSP Int CityMap
  deriving Show

type CityMap = M.Map City (M.Map City Weight)
type City = Int
type Weight = Int

instance GAData TSP where
  dataRange (TSP n _) = (0, n-1)
  repLength (TSP n _) = n
  fitness d@(TSP _ cm) order = allCities*pathExists*(maxFitness d - distance)
    where
      allCities = if allUnique order then 1 else 0
      paths = zip order $ tail order
      maybeDistance (s, e) = M.lookup s cm >>= M.lookup e
      maybeDistances = map maybeDistance paths
      distances = map (fromMaybe 0) maybeDistances
      distance = sum distances
      pathExists = if product distances > 0 then 1 else 0
  maxFitness (TSP n cm) = 1 + (n-1) * M.foldr (\m acc -> max acc $ M.foldr max 0 m) 0 cm
  -- we add 1 to the maxFitness so that "longest" paths get fitness 1, but invalid paths get fitness 0
  showSol _ order = show order
